/* Based on BSD last's sources and modified for amstat statistic server
 * Modifications Copyright (c) 1996 Andrew Maltsev <am@amsoft.ru>
 *
 * $Source: /src/CVSROOT/amstat/modules/calctime/amlast.c,v $
 *
 *    wtmp       .
*/

static char const rcsid[]=
   "$Id: amlast.c,v 1.6 1997/01/18 17:57:07 am Exp $";
/* $Log: amlast.c,v $
 * Revision 1.6  1997/01/18 17:57:07  am
 *     crash'     
 *    .      ,   .
 *  ,      (shutdown) 
 * .
 *
 * Revision 1.5  1996/12/02 11:09:15  am
 *      host'
 *
 * Revision 1.4  1996/11/22 17:32:46  am
 *   -T
 *
 * Revision 1.3  1996/11/22 17:20:02  am
 *    -      (-f)   
 *  ,        -  
 *    -  `still in'     ,
 *     .
 *
 *           crash
 *
 *   Linux
 *
 * Revision 1.2  1996/08/09 18:35:16  am
 *        
 * ,   .
 *
*/
#include <sys/param.h>
#include <sys/stat.h>

#include <fcntl.h>
#include <paths.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <utmp.h>
#ifndef linux
#include <err.h>
#else
#include <stdarg.h>
#endif

#define	COMPATIBILITY		/* To allow using `last v7' in place	*/
				/* of `last -t ttyv7'			*/

#define	NO	0				/* false/no */
#define	YES	1				/* true/yes */

static struct utmp buf[1024];			/* utmp read buffer */

typedef struct arg {
	char	*name;				/* argument */
#define	HOST_TYPE	-2
#define	TTY_TYPE	-3
#define	USER_TYPE	-4
	int	type;				/* type of arg */
	struct arg	*next;			/* linked list pointer */
} ARG;
ARG	*arglist;				/* head of linked list */

typedef struct ttytab {
	long	logout;				/* log out time		*/
	short	error;				/* error sign		*/
	short   closed;				/* closed tty?		*/
	char	tty[UT_LINESIZE + 1];		/* terminal name	*/
	struct ttytab	*next;			/* linked list pointer	*/
} TTY;
TTY	*ttylist;				/* head of linked list */

static long	crashtime=0,			/* current crash value */
                nowtime=0,			/* most recent time found */
		maxrec;				/* records to display */
static char	*file = _PATH_WTMP;		/* wtmp file */

static int	needcrtime=NO;

void	 addarg __P((int, char *));
TTY	*addtty __P((char *));
void	 hostconv __P((char *));
char	*ttyconv __P((char *));
int	 want __P((struct utmp *, int));
void	 wtmp __P((void));

#ifdef linux
void
err(int status, char const *fmt, ...)
{ va_list ap;
  va_start(ap,fmt);
  vfprintf(stderr,fmt,ap);
  fputc('\n',stderr);
  exit(status);
}
#endif

int
main(int argc, char *argv[])
{ extern int optind;
  extern char *optarg;
  int ch;
  char *p;

  nowtime = time(0);	/* Current time */

  maxrec = -1;
  while ((ch = getopt(argc, argv, "0123456789f:h:t:T:")) != EOF)
   switch (ch)
    { case 'T':			/* a way to set current time	*/
       nowtime=atol(optarg);	/* definitely not a best way	*/
       break;
/*
 * kludge: last was originally designed to take
 * a number after a dash.
 */
      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
       if (maxrec == -1)
        { p=argv[optind-1];
          if(p[0]=='-' && p[1]==ch&&!p[2])
           maxrec = atol(++p);
          else
           maxrec = atol(argv[optind] + 1);
          if(!maxrec)
           exit(0);
        }
       break;
      case 'f':
       file = optarg;
       nowtime = 0;			/* Will be set to file's most	*/
					/* recent time			*/
       break;
      case 'h':
       hostconv(optarg);
       addarg(HOST_TYPE, optarg);
       break;
      case 't':
       addarg(TTY_TYPE, ttyconv(optarg));
       break;
      case '?':
      default:
       fprintf(stderr,
"amlast -- shows wtmp information in computer readable form\n\n"
"usage: amlast [-#] [-f file] [-t tty] [-h hostname] [-T time] [user ...]\n");
       exit(1);
    }

  if(argc)
   { setlinebuf(stdout);
     for(argv+=optind; *argv; ++argv)
      {
#ifdef	COMPATIBILITY
        addarg(TTY_TYPE, ttyconv(*argv)); /* code to allow "last p5" to work */
#endif
        addarg(USER_TYPE, *argv);
      }
   }
  wtmp();
  exit(0);
}

/*     utmp
*/
static
void
print_ut(struct utmp *bp, long q, char *cmt)
{ struct tm *ptm;
  char *s;

#ifdef linux
  if(!strcmp(bp->ut_name,"LOGIN"))
   return;
#endif

/* Cleaning spaces in host name for 'FIDO 14400/V.42' logins from mgetty
*/
  for(s=bp->ut_host; *s; s++)
   if(*s==' ') *s='_';

# define sfill(x,y) ((*x) ? (x) : (y))
  printf("%-*.*s  %-*.*s %-*.*s "
        ,UT_NAMESIZE, UT_NAMESIZE, sfill(bp->ut_name,"unknown")
        ,UT_LINESIZE, UT_LINESIZE, sfill(bp->ut_line,"unknown")
        ,UT_HOSTSIZE, UT_HOSTSIZE, sfill(bp->ut_host,"localhost"));
# undef sfill

  ptm=localtime(&bp->ut_time);
  printf("%02d/%02d/%04d %02d:%02d:%02d "
        ,ptm->tm_mday,ptm->tm_mon+1,ptm->tm_year+1900
        ,ptm->tm_hour,ptm->tm_min,ptm->tm_sec);
  printf("%7ld",q<0 ? 0 : q);
  if(cmt || q<0)
   printf(" (%s)",cmt ? cmt : "will be more");
  puts("");
}
	
/* wtmp -- read through the wtmp file
*/
void
wtmp()
{ struct utmp *bp;			/* current structure */
  TTY *T;				/* tty list entry */
  struct stat stb;			/* stat of file for size */
  long bl;				/* time difference */
  int bytes=0, wfd;
  char *crmsg="error";

  if((wfd=open(file,O_RDONLY,0))<0 || fstat(wfd,&stb))
   err(1,"%s",file);

/*     
*/
  bl=(stb.st_size+sizeof(buf)-1)/sizeof(buf);

  time(&buf[0].ut_time);
  
  while(--bl>=0)
   { if(lseek(wfd,(off_t)(bl*sizeof(buf)),SEEK_SET)==-1 ||
  	(bytes=read(wfd,buf,sizeof(buf)))==-1)
      err(1, "%s", file);
     for(bp=buf+(bytes/sizeof(buf[0])-1); bp>=buf; --bp)
      { if(bp->ut_time>nowtime)
         nowtime=bp->ut_time+1;		/* When we feeded from file.. */

/* ~ -  crash  shutdown -  
 *
 *   ,     wtmp  
 *  ~/reboot    (     ,
 *       ).
 *
 *       ~/shutdown  
 *    .    
 *     reboot  .
 *
 *       ,   
 *     ~    
 *
*/
        if(bp->ut_line[0]=='~' && !bp->ut_line[1])
         { for(T=ttylist; T; T=T->next)     /* everybody just logged out */
            if(!T->closed)
             T->error=1;
           crashtime=bp->ut_time;
           needcrtime=YES;
	   crmsg=!strncmp(bp->ut_name,"shutdown",UT_NAMESIZE) ?
	         "shutdown" : "crash";
           if(want(bp,NO))
            { print_ut(bp,0,NULL);
	      if(maxrec!=-1 && !--maxrec)
	       return;
	    }
	   continue;
	 }

/* if the line is '{' or '|', date got set; see utmp(5) for more info.
*/
	if((bp->ut_line[0]=='{' || bp->ut_line[0]=='|') && !bp->ut_line[1])
	 { static time_t new_t;
	   if(want(bp,NO))
	    { print_ut(bp,0,bp->ut_line[0]=='{' ? "new time" : "old time");
	      if(maxrec && !--maxrec)
	       return;
	    }
	   if(bp->ut_line[0]=='{')
	    new_t=bp->ut_time;
	   else
            { for(T=ttylist; T; T=T->next)	/*      */
               { T->logout-=new_t-bp->ut_time;	/*   */
                 T->error=1;			/* ..        */
               }
              crmsg="time corr.";
              if(needcrtime)
               { crashtime-=new_t-bp->ut_time;
                 needcrtime=NO;
               }
            }
	   continue;
	 }

/*  crashtime
*/
        if(needcrtime)
         { crashtime=bp->ut_time+1;
           needcrtime=NO;
         }

/* ģ  
*/
	for(T=ttylist; T; T=T->next)
	 if(!strncmp(T->tty,bp->ut_line,UT_LINESIZE))
	  break;
	if(!T)
	 T=addtty(bp->ut_line);			/* add new one */

/* ,  
*/
	if(bp->ut_name[0] && want(bp,YES))
	 { time_t logout;
	   char *cmt=NULL;
	   if(!T->logout)
	    { cmt="still in";
	      logout=nowtime;
	    }
	   else
	    { if(T->error)
               { cmt=crmsg;
                 logout=crashtime;
               }
	      else
	       logout=T->logout;
	    }
	   print_ut(bp,logout-bp->ut_time,cmt);
	   if(maxrec!=-1 && !--maxrec)
	    return;
	 }

/*   
*/
	T->logout = bp->ut_time;
	T->error = 0;
	T->closed = !(*bp->ut_name);
      }
   }
}

/* want -- see if want this entry
*/
int
want(bp, check)
	struct utmp *bp;
	int check;
{
	ARG *step;

	if (check)
		/*
		 * when uucp and ftp log in over a network, the entry in
		 * the utmp file is the name plus their process id.  See
		 * etc/ftpd.c and usr.bin/uucp/uucpd.c for more information.
		 */
		if (!strncmp(bp->ut_line, "ftp", sizeof("ftp") - 1))
			bp->ut_line[3] = '\0';
		else if (!strncmp(bp->ut_line, "uucp", sizeof("uucp") - 1))
			bp->ut_line[4] = '\0';
	if (!arglist)
		return (YES);

	for (step = arglist; step; step = step->next)
		switch(step->type) {
		case HOST_TYPE:
			if (!strncasecmp(step->name, bp->ut_host, UT_HOSTSIZE))
				return (YES);
			break;
		case TTY_TYPE:
			if (!strncmp(step->name, bp->ut_line, UT_LINESIZE))
				return (YES);
			break;
		case USER_TYPE:
			if (!strncmp(step->name, bp->ut_name, UT_NAMESIZE))
				return (YES);
			break;
	}
	return (NO);
}

/*
 * addarg --
 *	add an entry to a linked list of arguments
 */
void
addarg(type, arg)
	int type;
	char *arg;
{
	ARG *cur;

	if (!(cur = (ARG *)malloc((u_int)sizeof(ARG))))
		err(1, "malloc failure");
	cur->next = arglist;
	cur->type = type;
	cur->name = arg;
	arglist = cur;
}

/*
 * addtty --
 *	add an entry to a linked list of ttys
 */
TTY *
addtty(ttyname)
	char *ttyname;
{ TTY *cur,*tmp;

  if(!(cur=(TTY *)malloc((u_int)sizeof(TTY))))
   err(1, "malloc failure");
  cur->next = ttylist;
  cur->logout = crashtime;
  cur->closed = 0;

/*      ,     
*/
  cur->error = 0;
  for(tmp=ttylist; tmp; tmp=tmp->next)
   if(tmp->error)
    { cur->error=1;
      break;
    }

  memmove(cur->tty, ttyname, UT_LINESIZE);
  return (ttylist=cur);
}

/*
 * hostconv --
 *	convert the hostname to search pattern; if the supplied host name
 *	has a domain attached that is the same as the current domain, rip
 *	off the domain suffix since that's what login(1) does.
 */
void
hostconv(arg)
	char *arg;
{
	static int first = 1;
	static char *hostdot, name[MAXHOSTNAMELEN];
	char *argdot;

	if (!(argdot = strchr(arg, '.')))
		return;
	if (first) {
		first = 0;
		if (gethostname(name, sizeof(name)))
			err(1, "gethostname");
		hostdot = strchr(name, '.');
	}
	if (hostdot && !strcasecmp(hostdot, argdot))
		*argdot = '\0';
}

/*
 * ttyconv --
 *	convert tty to correct name.
 */
char *
ttyconv(arg)
	char *arg;
{
	char *mval;

	/*
	 * kludge -- we assume that all tty's end with
	 * a two character suffix.
	 */
	if (strlen(arg) == 2) {
		/* either 6 for "ttyxx" or 8 for "console" */
		if (!(mval = malloc((u_int)8)))
			err(1, "malloc failure");
		if (!strcmp(arg, "co"))
			(void)strcpy(mval, "console");
		else {
			(void)strcpy(mval, "tty");
			(void)strcpy(mval + 3, arg);
		}
		return (mval);
	}
	if (!strncmp(arg, _PATH_DEV, sizeof(_PATH_DEV) - 1))
		return (arg + 5);
	return (arg);
}
